#ifndef _zrp_table_h_
#define _zrp_table_h_

#include <config.h>
#include <agent.h>
#include <packet.h>
#include <ip.h>
#include <delay.h>
#include <scheduler.h>
#include <queue.h>
#include <trace.h>
#include <arp.h>
#include <ll.h>
#include <mac.h>
#include <priqueue.h>
#include <delay.h>
#include <random.h>
#include <object.h>
#include <route.h>
#include "zrp.h"

//typedef int32_t Query_ID;
typedef int Query_ID;
typedef int BOOLEAN;
typedef double Time;



#define LINKSTATE_EXPIRATION 25

#define MAX_NODES_PER_ZONE 50
#define MAX_OUTER_ROUTES 50

class ZRPAgent; // pre-declaration

//===========================================================================
// IERP Cache Table:
// A single entry in the outgoing IERP requests table, has an expiry.
//===========================================================================
struct IERPCacheEntry {
  
  Query_ID qid_;
  Packet *pkt_;
  Time expiry_;
};

struct rtable_node; // pre-declaration needed for next struct

//===========================================================================
// Query Detect Table:
// A single entry in the QD table, has an expiry.
//===========================================================================
struct QueryDetectEntry {  

  nsaddr_t originid_;
  Query_ID qid_;
  Time expiry_;
  rtable_node *next_; // list of covered nodes

};

//===========================================================================
// Neighbor Table:
// Neighbor table is a linked list of these.
//===========================================================================
struct neighbor_struct 
{
  nsaddr_t addr_;
  neighbor_struct* next_;
  Time expiry_;
  Time lastack_;
  
};

//===========================================================================
// Linkstate Table:
// Link state atomic data, contains state in isup_ (up or down) ans seq.
//===========================================================================
class link_struct {
public:

  link_struct() : seq_(0), isup_(FALSE) {//, next_(NULL) {
    bzero(this, sizeof(link_struct));
  }

  link_struct(nsaddr_t srcid, nsaddr_t destid, BOOLEAN valid,link_struct* next)
    {
      src_=(srcid);
      dest_=(destid);
      isup_=(valid);
      //      next_=(next);
      seq_=0;
    }

  nsaddr_t src_;   // 32 bits
  nsaddr_t dest_;  // 32 bits

  int flag_; // flag used for Dijstkra's algorithm
  int isup_;
  int seq_;  // sequence number
  Time expiry_;
    
  //  Time timestamp;
  
  link_struct* next_;
};

//===========================================================================
// Linkstate Table:
// Class for storing Link States in node.
//===========================================================================
class LinkStateTable {   
 public:

  LinkStateTable() : lshead_(NULL) { }

  nsaddr_t  my_address_; // debug
  
  link_struct* lshead_;

  link_struct* last() ;

  void add(link_struct *link);  // quietly ignore errors ;)

  int remove(link_struct *link);

  BOOLEAN update(nsaddr_t src, nsaddr_t dest, 
		 int state, int seq, int *seq_match);

  int exist(link_struct *link);
  
  void print_links();
  void purge_links();
  

};

//===========================================================================
// Route Table:
// A single data struct naming a node, used for storing node addresses
// in various tables.
//===========================================================================
struct rtable_node {
  nsaddr_t id_;
  int numhops_; // number of hops from me
  rtable_node* next_;
  rtable_node() : id_(0), next_(NULL) { }
  rtable_node(nsaddr_t newid) : id_(newid), next_(NULL) { }
};

//===========================================================================
// Route Table:
// This struct is a head pointer for a route entry, has expiry timestamp.
// Route is at next_ .
//===========================================================================
struct rtable_entry {
  nsaddr_t id_;
  Time expiry_;
  int numhops_; // number of hops from me
  rtable_node* next_; // route is here
  rtable_entry() : id_(0), next_(NULL), expiry_(0), numhops_(0)  { }
  rtable_entry(nsaddr_t newid) : id_(newid), next_(NULL), expiry_(0),
    numhops_(0) { }
};

//===========================================================================
// Route Table:
// Class includes a tag (agent_) to agent it resides in to access friendly
// data and methods.
//===========================================================================
class RoutingTable {
  public:

  RoutingTable(ZRPAgent* a)  { // init
    num_entries_=(0);
    agent_ = a;
  }
  
  ZRPAgent* agent_;
  
  nsaddr_t  my_address_;

  int num_entries_;

  // Tables
  rtable_node* route[MAX_NODES_PER_ZONE+1]; // Last ptr is always null
  rtable_entry* outer_route[MAX_OUTER_ROUTES]; // Last ptr is always null
  rtable_node* periphnodes_;
  LinkStateTable *linkstatetable;

  // Methods    
  void RoutingTable::compute_routes(nsaddr_t myadd, int radius);
  void erase();
  void print();
  // returns pointer to entry to node's route
  int get_entry(nsaddr_t node_id);
  int route_exists(nsaddr_t id);

  // drop from link table, add node to RouteTable
  void add_drop(link_struct* ls, int i);

  // Finds Next Hop On route to given,
  //returns FALSE if doesn't exist
  int next_hop_to(nsaddr_t node) { }

  // returns route entry index for given node 
  int node_exists(nsaddr_t nodei);

  void empty_links(link_struct *lst);

  //  int rmax; // the number of hops to the farthest node in my LRZ
};

#endif
